from marshmallow import fields

from prefect.utilities.serialization import JSONCompatible, OneOfSchema, ObjectSchema
from prefect.run_configs import KubernetesRun


class RunConfigSchemaBase(ObjectSchema):
    labels = fields.List(fields.String())


class KubernetesRunSchema(RunConfigSchemaBase):
    class Meta:
        object_class = KubernetesRun

    job_template_path = fields.String(allow_none=True)
    job_template = JSONCompatible(allow_none=True)
    image = fields.String(allow_none=True)
    env = fields.Dict(keys=fields.String(), allow_none=True)
    cpu_limit = fields.String(allow_none=True)
    cpu_request = fields.String(allow_none=True)
    memory_limit = fields.String(allow_none=True)
    memory_request = fields.String(allow_none=True)


class RunConfigSchema(OneOfSchema):
    type_schemas = {
        "KubernetesRun": KubernetesRunSchema,
    }



import json
from typing import Any

from prefect import Task
from prefect.utilities.tasks import defaults_from_attrs


class CreateIssueComment(Task):
    """
    Task for creating a comment on the specified GitHub issue using the v3 version of the
    GitHub REST API.

    Args:
        - repo (str, optional): the name of the repository to create the comment in; must be
            provided in the form `organization/repo_name`; can also be provided to the `run`
            method
        - issue_number (int, optional): the ID of the issue to create the comment on.
        - body (str, optional): the body of the comment; can also be provided to the `run` method
        - **kwargs (Any, optional): additional keyword arguments to pass to the standard Task
            init method

    References:
        - See: [Create an issue
            comment](https://developer.github.com/v3/issues/comments/#create-an-issue-comment)
    """

    def __init__(
        self,
        repo: str = None,
        issue_number: int = None,
        body: str = None,
        **kwargs: Any
    ):
        self.repo = repo
        self.issue_number = issue_number
        self.body = body
        super().__init__(**kwargs)

    @defaults_from_attrs("repo", "issue_number", "body")
    def run(
        self,
        repo: str = None,
        issue_number: int = None,
        body: str = None,
        token: str = None,
    ) -> None:
        """
        Run method for this Task. Invoked by calling this Task after initialization within a
        Flow context, or by using `Task.bind`.

        Args:
            - repo (str, optional): the name of the repository to create the comment in; must be
                provided in the form `organization/repo_name`; can also be provided to the `run`
                method
            - issue_number (int, optional): the ID of the issue to create the comment on.
            - body (str, optional): the body of the comment; can also be provided to the `run` method
            - token (str): a GitHub API token

        Raises:
            - ValueError: if a `repo` or an `issue_number` was never provided
            - HTTPError: if the POST request returns a non-200 status code

        Returns:
            - None
        """
        if repo is None:
            raise ValueError("A GitHub repository must be provided.")

        if issue_number is None:
            raise ValueError("An issue number must be provided")

        # 'import requests' is expensive time-wise, we should do this just-in-time to keep
        # the 'import prefect' time low
        import requests

        url = "https://api.github.com/repos/{}/issues/{}/comments".format(
            repo, issue_number
        )
        headers = {
            "AUTHORIZATION": "token {}".format(token),
            "Accept": "application/vnd.github.v3+json",
        }
        issue = {"body": body}

        # send the request
        resp = requests.post(url, data=json.dumps(issue), headers=headers)
        resp.raise_for_status()



import importlib
from typing import TYPE_CHECKING

import prefect

if TYPE_CHECKING:
    from prefect.core.flow import Flow  # pylint: disable=W0611


def get_flow_image(flow: "Flow") -> str:
    """
    Retrieve the image to use for this flow deployment. Will start by looking for
    an `image` value in the flow's `environment.metadata`. If not found then it will fall
    back to using the `flow.storage`.

    Args:
        - flow (Flow): A flow object

    Returns:
        - str: a full image name to use for this flow run

    Raises:
        - ValueError: if deployment attempted on unsupported Storage type and `image` not
            present in environment metadata
    """
    environment = flow.environment
    if hasattr(environment, "metadata") and environment.metadata.get("image"):
        return environment.metadata.get("image", "")
    else:
        storage = flow.storage
        if not isinstance(storage, prefect.environments.storage.Docker):
            raise ValueError(
                f"Storage for flow run {flow.name} is not of type Docker and "
                f"environment has no `image` attribute in the metadata field."
            )

        return storage.name


def extract_flow_from_file(
    file_path: str = None, file_contents: str = None, flow_name: str = None
) -> "Flow":
    """
    Extract a flow object from a file.

    Args:
        - file_path (str, optional): A file path pointing to a .py file containing a flow
        - file_contents (str, optional): The string contents of a .py file containing a flow
        - flow_name (str, optional): A specific name of a flow to extract from a file.
            If not set then the first flow object retrieved from file will be returned.

    Returns:
        - Flow: A flow object extracted from a file

    Raises:
        - ValueError: if both `file_path` and `file_contents` are provided or neither are.
    """
    if file_path and file_contents:
        raise ValueError("Provide either `file_path` or `file_contents` but not both.")

    if not file_path and not file_contents:
        raise ValueError("Provide either `file_path` or `file_contents`.")

    # Read file contents
    if file_path:
        with open(file_path, "r") as f:
            contents = f.read()

    # Use contents directly
    if file_contents:
        contents = file_contents

    # Load objects from file into dict
    exec_vals = {}  # type: ignore
    exec(contents, exec_vals)

    # Grab flow name from values loaded via exec
    for var in exec_vals:
        if isinstance(exec_vals[var], prefect.Flow):
            if flow_name and exec_vals[var].name == flow_name:
                return exec_vals[var]
            elif not flow_name:
                return exec_vals[var]

    raise ValueError("No flow found in file.")


def extract_flow_from_module(module_str: str, flow_name: str = None) -> "Flow":
    """
    Extract a flow object from a python module.

    Args:
        - module_str (str): A module path pointing to a .py file containing a flow.
            For example, 'myrepo.mymodule.myflow' where myflow.py contains the flow.
        - flow_name (str, optional): A specific name of a flow to extract from a file.
            If not set then the first flow object retrieved from file will be returned.
            The "first" flow object will be based on the dir(module) ordering, which
            is alphabetical and capitalized letters come first.

    Returns:
        - Flow: A flow object extracted from a file
    """

    # load the module
    module = importlib.import_module(module_str)

    # if flow_name is specified, grab it from the module
    if flow_name:
        return getattr(module, flow_name)
    # otherwise loop until we get a Flow object
    else:
        for var in dir(module):
            if isinstance(getattr(module, var), prefect.Flow):
                return getattr(module, var)

    raise ValueError("No flow found in module.")
